"""generates wrapper methods for several different vector types, which would be
tedious to do by hand"""

# must work in CPython 2.7 and 3.0+

from __future__ import print_function

import json
import re

from collections import namedtuple


# 64 is intentionally omitted (not supported)
widths = [128,256,512]

FLOAT = 1
INTEGER = 2

CType = namedtuple('CType','name size code float')
types = {}

for t in [
    ('int8_t',8,'i8',INTEGER),
#    ('uint8_t',8,'u8',INTEGER),
    ('int16_t',16,'i16',INTEGER),
#    ('uint16_t',16,'u16',INTEGER),
    ('int32_t',32,'i32',INTEGER),
#    ('uint32_t',32,'u32',INTEGER),
    ('int64_t',64,'i64',INTEGER),
#    ('uint64_t',64,'u64',INTEGER),
    ('float',32,'s',FLOAT),
    ('double',64,'d',FLOAT)]:
    types[t[0]] = CType(*t)


binary_op = """        FORCE_INLINE {wrap_type} operator{extra}({wrap_type} b) const {{
            return {intr}(this->data.p,b.data.p);
        }}
        FORCE_INLINE {wrap_type} &operator{extra}=({wrap_type} b) {{
            this->data.p = {intr}(this->data.p,b.data.p);
            return *this;
        }}
"""
cmp_op = """        FORCE_INLINE mask operator{extra[0]}({wrap_type} b) const {{
            return mask({intr}(this->data.p,b.data.p{extra[1]}));
        }}
"""
unary_meth = """        FORCE_INLINE {wrap_type} {extra}() const {{
            return {intr}(this->data.p);
        }}
"""
reduce_meth = """        FORCE_INLINE {item_type} {extra}() const {{
            return {intr}(this->data.p);
        }}
"""

have_vec_reduce_add = '        static constexpr bool has_vec_reduce_add = true;\n'
reduce_add_meth = reduce_meth + have_vec_reduce_add
reduce_add_meth_b = """        FORCE_INLINE single<{item_type}> reduce_add() const {{
            auto r = {intr}(this->data.p,this->data.p);{more_steps}
            return {intr}(r,r);
        }}
""" + have_vec_reduce_add

load_meth = """        static FORCE_INLINE {wrap_type} {extra}(const {item_type} *source) {{
            return {intr}(reinterpret_cast<{iargs[0][type]}>(source));
        }}
"""

store_meth = """        FORCE_INLINE void {extra}({item_type} *dest) const {{
            {intr}(reinterpret_cast<{iargs[0][type]}>(dest),this->data.p);
        }}
"""

set1_meth = """        static FORCE_INLINE {wrap_type} {extra}({item_type} x) {{
            return {intr}(x);
        }}
"""

broadcast_meth = """        static FORCE_INLINE {wrap_type} repeat(single<{item_type}> x) {{
            return {intr}(x.data);
        }}
"""
shuffle_meth = """        static FORCE_INLINE {wrap_type} repeat(single<{item_type}> x) {{
            return {intr}(x.data{extra},_MM_SHUFFLE(0,0,0,0));
        }}
"""

setzero_meth = """        static FORCE_INLINE {wrap_type} {extra}() {{
            return {intr}();
        }}
"""

cast_meth = """        explicit FORCE_INLINE {wrap_type}({wrap_type_src} b) {{
            this->data.p = {intr}(b.data.p);
        }}
"""
cast_meth_dec = """        explicit {wrap_type}({wrap_type_src} b);
"""
cast_meth_def = """    FORCE_INLINE {wrap_type}::{wrap_type}({wrap_type_src} b) {{
        this->data.p = {intr}(b.data.p);
    }}
"""

binary_func = """    FORCE_INLINE {wrap_type} {extra}({wrap_type} a,{wrap_type} b) {{
        return {intr}(a.data.p,b.data.p);
    }}
"""
cmp_func = """    FORCE_INLINE {wrap_type}::mask {extra[0]}({wrap_type} a,{wrap_type} b) {{
        return {wrap_type}::mask({intr}(a.data.p,b.data.p{extra[1]}));
    }}
"""

mask_func = """    FORCE_INLINE mask{ssize}_v_{psize} {extra}({args_a}) {{
        return
"""

def print_mask_function(output,swidth,pwidth,func,args,body,fallback,explicit_cast=True):
    print('    FORCE_INLINE mask{0}_v_{1} {2}({3}) {{\n        return '.format(
        swidth,
        pwidth,
        func,
        ','.join('mask{0}_v_{1} {2}'.format(swidth,pwidth,chr(ord('a')+i)) for i in range(args))),end='',file=output)
    if explicit_cast:
        print('mask{0}_v_{1}('.format(swidth,pwidth),end='',file=output)
    
    if fallback:
        print(
            '\n    ',
            ifdef_support(fallback[0]),
            '\n            ',
            body,
            '\n    #else\n            '
            ,sep='',end='',file=output)
        
        body = fallback[1]
    
    print(body,end='',file=output)

    if fallback:
        print('\n    #endif\n        ',end='',file=output)
    
    if explicit_cast:
        print(')',end='',file=output)
    print(';\n    }',file=output)


def mm_prefix(width):
    return '_mm{0}'.format(width if width > 128 else '')

def generic_intrinsic_name(type,width,base,scalar=False):
    return '{0}_{1}_{2}{3}{4}'.format(
        mm_prefix(width),
        base,
        '' if width == 64 or type.float == FLOAT else 'e',
        's' if scalar else 'p',
        type.code)

def int_log2(x):
    r = 0
    x >>= 1
    while x:
        r += 1
        x >>= 1
    return r

class FTransform(object):
    def __init__(self,base_name,template,extra=None,support=FLOAT|INTEGER,prefer=None,scalar=False):
        self.base_name = base_name
        self.template = template
        self.extra = extra or base_name
        self.support = support
        self.prefer = prefer
        self.scalar = scalar
        
    def supported(self,type,width):
        return self.support & type.float and (width != 512 or type.size >= 32)
        
    def intrinsic(self,type,width):
        return generic_intrinsic_name(type,width,self.base_name,self.scalar)
    
    def output(self,type,width,intr):
        return self.template.format(
            intr=intr['name'],
            iargs=intr['parameters'],
            wrap_type=wrap_type(type,width),
            base_type=base_type(type,width),
            item_type=type.name,
            extra=self.extra)
    
    def antirequisite(self,type,width,intrinsics):
        if self.prefer:
            if self.prefer.supported(type,width):
                r = intrinsics.get(self.prefer.intrinsic(type,width))
                if r: return r['tech']
            return self.prefer.antirequisite(type,width,intrinsics)
        return None
    
class ShuffleTransform(FTransform):
    def __init__(self,prefer=None):
        super(ShuffleTransform,self).__init__('shuffle',shuffle_meth,prefer=prefer)
    
    def supported(self,type,width):
        return width == 128 and type.size >= 32

    def output(self,type,width,intr):
        return self.template.format(
            intr=intr['name'],
            wrap_type=wrap_type(type,width),
            item_type=type.name,
            extra=',x.data' if type.float == FLOAT else '')

class BroadcastTransform(FTransform):
    def __init__(self):
        super(BroadcastTransform,self).__init__(None,broadcast_meth)
    
    def intrinsic(self,type,width):
        return generic_intrinsic_name(type,width,'broadcast' + {'int8_t':'b','int16_t':'w','int32_t':'d','int64_t':'q','float':'ss','double':'sd'}[type.name]) 

class ReduceAddBTransform(FTransform):
    def __init__(self,prefer=None):
        super(ReduceAddBTransform,self).__init__('hadd',reduce_add_meth_b,prefer=prefer)
    
    def output(self,type,width,intr):
        ms = int_log2(width//type.size) - 2
        if ms:
            ms = '\n            r = {0}(r,r);'.format(intr['name']) * ms
        else:
            ms = ''
        
        return self.template.format(
            intr=intr['name'],
            wrap_type=wrap_type(type,width),
            item_type=type.name,
            more_steps=ms)

class AltIFTransform(FTransform):
    def intrinsic(self,type,width):
        if type.float == INTEGER:
            return '{0}_{1}_si{2}'.format(
                mm_prefix(width),
                self.base_name,
                width)
        
        return super(AltIFTransform,self).intrinsic(type,width)

class CmpFTransform(FTransform):
    def __init__(self,op,template,code=None):
        self.op = op
        self.code = code
        super(CmpFTransform,self).__init__(None,template)
        
    def intrinsic(self,type,width):
        base = 'cmp'
        if width < 256 or type.float != FLOAT: base += self.code
        
        r = generic_intrinsic_name(type,width,base)
        
        if width == 512:
            r += '_mask'
        
        return r
    
    def output(self,type,width,intr):
        arg3 = ''
        if width >= 256 and type.float == FLOAT:
            arg3 = ',_CMP_{0}_UQ'.format(self.code.upper())

        return self.template.format(
            intr=intr['name'],
            iargs=intr['parameters'],
            wrap_type=wrap_type(type,width),
            base_type=base_type(type,width),
            item_type=type.name,
            extra=(self.op,arg3))

class RepeatTransform(FTransform):
    def intrinsic(self,type,width):
        r = super(RepeatTransform,self).intrinsic(type,width)
        if type is types['int64_t'] and width < 512:
            r += 'x'
        return r

class CastTransform(object):
    def supported(self,type,width,src_width,inline):
        return not (width == src_width or
            (width < src_width and not inline) or
            ((width == 512 or src_width == 512) and type.size < 32))
    
    def intrinsic(self,type,width,src_width,inline):
        code = 'si'
        if type is types['double']: code = 'pd'
        elif type is types['float']: code = 'ps'
        return '{0}_cast{1}{2}_{1}{3}'.format(
            mm_prefix(max(width,src_width)),
            code,
            src_width,
            width)
    
    def output(self,type,width,src_width,intr,inline):
        tmpl = cast_meth
        if width > src_width:
            tmpl = cast_meth_dec if inline else cast_meth_def
        else:
            assert inline
        
        return tmpl.format(
            intr=intr['name'],
            wrap_type=wrap_type(type,width),
            wrap_type_src=wrap_type(type,src_width))

class MixedFTransformAdapter(FTransform):
    def __init__(self,src_width,base):
        self.src_width = src_width
        self.base = base
        super(MixedFTransformAdapter,self).__init__('',None)
        
    def intrinsic(self,type,width):
        return CastTransform().intrinsic(type,width,self.src_width,True)
    
    def output(self,type,width,intr):
        return CastTransform().output(type,width,self.src_width,intr,True)


mixed_method_transforms = [
    CastTransform()]

reduce_add = FTransform('reduce_add',reduce_add_meth)
repeat_a = BroadcastTransform()

method_transforms = [MixedFTransformAdapter(w,mmt) for w in widths for mmt in mixed_method_transforms] + [
    FTransform('add',binary_op,'+'),
    FTransform('sub',binary_op,'-'),
    FTransform('mul',binary_op,'*',support=FLOAT),
    FTransform('div',binary_op,'/',support=FLOAT),
    AltIFTransform('and',binary_op,'&'),
    AltIFTransform('or',binary_op,'|'),
    AltIFTransform('xor',binary_op,'^'),
    CmpFTransform('==',cmp_op,'eq'),
    CmpFTransform('!=',cmp_op,'neq'),
    CmpFTransform('>',cmp_op,'gt'),
    CmpFTransform('>=',cmp_op,'ge'),
    CmpFTransform('<',cmp_op,'lt'),
    CmpFTransform('<=',cmp_op,'le'),
    FTransform('sqrt',unary_meth,support=FLOAT),
    FTransform('rsqrt',unary_meth,support=FLOAT),
    FTransform('abs',unary_meth),
    FTransform('ceil',unary_meth,support=FLOAT),
    FTransform('floor',unary_meth,support=FLOAT),
    reduce_add,
    ReduceAddBTransform(prefer=reduce_add),
    FTransform('reduce_max',reduce_meth),
    FTransform('reduce_min',reduce_meth),
    AltIFTransform('load',load_meth),
    AltIFTransform('loadu',load_meth),
    AltIFTransform('store',store_meth),
    AltIFTransform('storeu',store_meth),
    repeat_a,
    ShuffleTransform(prefer=repeat_a),
    RepeatTransform('set1',set1_meth,'repeat'),
    AltIFTransform('setzero',setzero_meth,'zeros')]

function_transforms = [
    FTransform('max',binary_func),
    FTransform('min',binary_func),
    AltIFTransform('andnot',binary_func,'and_not'),
    CmpFTransform('cmp_nlt',cmp_func,'nlt'),
    CmpFTransform('cmp_nle',cmp_func,'nle'),
    CmpFTransform('cmp_ngt',cmp_func,'ngt'),
    CmpFTransform('cmp_nge',cmp_func,'nge')]


def data_args(args):
    return ['{0}.data'.format(chr(ord('a')+i)) for i in range(args)]

def ideal_args(swidth,pwidth,args):
    r = data_args(args)
    if swidth >= 32: r = ['reinterpret_cast<__m{0}i>({1})'.format(pwidth,a) for a in r]
    return ','.join(r)

def avx512_mask_body(swidth,base,args):
    r = '_mm512_k{0}({1})'.format(base,','.join(data_args(args)))
    if swidth != 32: r = 'static_cast<__mmask{0}>({1})'.format(512/swidth,r)
    return r

def common_mask_function(output,intrinsics,swidth,pwidth,type_req,suffix,name,intr_base):
    fallback = None
    
    if pwidth == 512:
        body = avx512_mask_body(swidth,intr_base,2)
    else:
        intr = intrinsics['{0}_{1}_si{2}'.format(mm_prefix(pwidth),intr_base,pwidth)]

        body = '{0}_{1}_si{2}({3})'.format(mm_prefix(pwidth),intr_base,pwidth,ideal_args(swidth,pwidth,2))
        
        if intr['tech'] not in IMPLIED_REQ[type_req]:
            assert swidth >= 32
            fallback = intr['tech'],'{0}_{1}_{2}(a.data,b.data)'.format(mm_prefix(pwidth),intr_base,suffix)
        
    print_mask_function(output,swidth,pwidth,name,2,body,fallback)

def get_eq_intr(intrinsics,pwidth,swidth):
    intr = intrinsics.get('{0}_cmpeq_epi{1}'.format(mm_prefix(pwidth),swidth))
    if intr: return intr,True
    
    return intrinsics['{0}_cmp_epi{1}'.format(mm_prefix(pwidth),swidth)],False

def not_mask_function(output,intrinsics,swidth,pwidth,type_req,suffix):
    fallback = None
    
    if pwidth == 512:
        body = avx512_mask_body(swidth,'not',1)
    else:
        intr,in_name = get_eq_intr(intrinsics,pwidth,swidth)

        body = '{0}_cmp{1}_epi{2}({3},{0}_setzero_si{4}(){5})'.format(
            mm_prefix(pwidth),
            ['','eq'][in_name],
            swidth,
            ideal_args(swidth,pwidth,1),
            pwidth,
            [',_CMP_EQ_UQ',''][in_name])
        
        if intr['tech'] not in IMPLIED_REQ[type_req]:
            assert swidth >= 32
            fallback = [intr['tech'],'{0}_cmp{1}_{2}(a.data,a.data{3})'.format(
                mm_prefix(pwidth),
                ['','unord'][in_name],
                suffix,
                [',_CMP_UNORD_Q',''][in_name])]
        
    print_mask_function(output,swidth,pwidth,'operator!',1,body,fallback)

def xnor_mask_function(output,intrinsics,swidth,pwidth,type_req,suffix):
    fallback = None
    
    if pwidth == 512:
        body = avx512_mask_body(swidth,'xnor',2)
    else:
        intr,in_name = get_eq_intr(intrinsics,pwidth,swidth)
        
        body = 'mask{0}_v_{1}({2}_cmp{3}_epi{0}({4}{5}))'.format(
            swidth,
            pwidth,
            mm_prefix(pwidth),
            ['','eq'][in_name],
            ideal_args(swidth,pwidth,2),
            [',_CMP_EQ_UQ',''][in_name])
        
        if intr['tech'] not in IMPLIED_REQ[type_req]:
            assert swidth >= 32
            fallback = intr['tech'],'!l_xor(a,b)'
        
    print_mask_function(output,swidth,pwidth,'l_xnor',2,body,fallback,pwidth == 512)


def macroize(x):
    return ''.join(c if c.isalnum() else '_' for c in x)

def wrap_type(type,width):
    return '{0}_v_{1}'.format(type.name[0:-2] if type.name.endswith('_t') else type.name,width)

def base_type(type,width):
    suffix = ''
    if type is types['double']: suffix = 'd'
    elif type.float == INTEGER: suffix = 'i'
    return '__m{0}{1}'.format(width,suffix)

def ifdef_support(req,antireq=None,indent=0):
    indent = '    '*indent
    if req and antireq:
        return '{0}#if defined(SUPPORT_{1}) && !defined(SUPPORT_{2})'.format(indent,macroize(req),macroize(antireq))

    extra = ''
    if not req:
        assert antireq
        extra = 'n'
        req = antireq
    return '{0}#if{1}def SUPPORT_{2}'.format(indent,extra,macroize(req))

def print_dependent(d_list,file,indent=0):
    for req,mt_list in d_list.items():
        print(ifdef_support(req[0],req[1],indent=indent),file=file)
            
        for trans,intr,type,width in mt_list:
            print(trans.output(type,width,intr),file=file)
            
        print('    ' * indent + '#endif',file=file)


IMPLIED_REQ = {}
rs = set()
for r in ['SSE','SSE2','SSE3','SSSE3','SSE4.1','SSE4.2','AVX','AVX2','AVX-512']:
    rs.add(r)
    IMPLIED_REQ[r] = frozenset(rs)
del rs


meth_search = re.compile(r'^\s*//\s*\[\[\[(\w+)-(\d+)(?:-(\d+)|)\]\]\]\s*$')
func_search = re.compile(r'^\s*//\s*\[\[\[functions\]\]\]\s*$')
mask_search = re.compile(r'^\s*//\s*\[\[\[mask-(\d+)-(\d+)\]\]\]\s*$')


def generate(data_file,template_file,output_file):
    with open(data_file,'r') as input:
        data = json.load(input)

    intrinsics = {}
    for item in data:
        intrinsics[item['name']] = item

    del data


    template = open(template_file,'r')
    output = open(output_file,'w')

    print('// THIS FILE WAS GENERATED BY generate_simd.py. DO NOT EDIT IT DIRECTLY.\n\n',file=output)

    for line in template:
        m = mask_search.match(line)
        if m:
            swidth = int(m.group(1),10)
            pwidth = int(m.group(2),10)
            
            if pwidth == 512:
                type_req = 'AVX-512'
            elif pwidth == 256:
                type_req = 'AVX2' if swidth < 32 else 'AVX'
            else:
                assert pwidth == 128
                type_req = 'SSE' if swidth == 32 else 'SSE2'
            
            if swidth == 64:
                suffix = 'pd'
            elif swidth == 32:
                suffix = 'ps'
            else:
                suffix = 'epi{0}'.format(swidth)
            
            common_mask_function(output,intrinsics,swidth,pwidth,type_req,suffix,'operator&&','and')
            common_mask_function(output,intrinsics,swidth,pwidth,type_req,suffix,'operator||','or')
            not_mask_function(output,intrinsics,swidth,pwidth,type_req,suffix)
            common_mask_function(output,intrinsics,swidth,pwidth,type_req,suffix,'l_andn','andn' if pwidth == 512 else 'andnot')
            common_mask_function(output,intrinsics,swidth,pwidth,type_req,suffix,'l_xor','xor')
            xnor_mask_function(output,intrinsics,swidth,pwidth,type_req,suffix)
            continue
        
        m = meth_search.match(line)
        if m:
            type = types[m.group(1)]
            width = int(m.group(2),10)
            assert width in widths
            
            if m.group(3):
                width2 = int(m.group(3),10)
                assert width2 in widths
                
                for mt in mixed_method_transforms:
                    if not mt.supported(type,width,width2,False): continue
                
                    intr_name = mt.intrinsic(type,width,width2,False)
                    intr = intr_name and intrinsics.get(intr_name)
                    if intr:
                        print(mt.output(type,width,width2,intr,False),file=output)
            else:
                if width == 512:
                    assert type.size >= 32
                    type_req = 'AVX-512'
                elif type is types['float']:
                    type_req = {128:'SSE',256:'AVX'}[width]
                elif type is types['double']:
                    type_req = {128:'SSE2',256:'AVX'}[width]
                elif width == 256:
                    type_req = 'AVX2'
                else:
                    assert width == 128
                    type_req = 'SSE2'
                
                dependent = {}
                for mt in method_transforms:
                    if mt.supported(type,width):
                        intr = intrinsics.get(mt.intrinsic(type,width))
                        if intr:
                            req = intr['tech']
                            antireq = mt.antirequisite(type,width,intrinsics)

                            if antireq is None and req in IMPLIED_REQ[type_req]:
                                print(mt.output(type,width,intr),file=output)
                            elif antireq is None or antireq not in IMPLIED_REQ[type_req]:
                                rs = (None if req in IMPLIED_REQ[type_req] else req,antireq)
                                if rs not in dependent:
                                    dependent[rs] = []
                                dependent[rs].append((mt,intr,type,width))
                        
                print_dependent(dependent,output,1)
            continue

        m = func_search.match(line)
        if m:
            dependent = {}
            for type in types.values():
                for width in widths:
                    for ft in function_transforms:
                        if not ft.supported(type,width): continue
                        
                        intr = intrinsics.get(ft.intrinsic(type,width))
                        if intr:
                            rs = (intr['tech'],ft.antirequisite(type,width,intrinsics))
                            if rs not in dependent:
                                dependent[rs] = []
                            dependent[rs].append((ft,intr,type,width))
                        
            print_dependent(dependent,output)
            continue

        print(line,end='',file=output)
                
    template.close()
    output.close()





import weakref
import pygame

import ntracer.render


def channels_from_surface(surface):
    """Create a list of :py:class:`.render.Channel` objects that match the
    pixel format of ``surface``.
    
    Note: indexed color modes (all 8-bit color modes are indexed) are not
    supported.
    
    :param surface: An instance of `pygame.Surface
        <http://www.pygame.org/docs/ref/surface.html#pygame.Surface>`_.
    
    """
    channels = []
    
    bs = surface.get_bytesize()
    if bs == 1:
        raise TypeError('indexed color modes are not supported')
    
    def_shift = (bs - 1) * 8
    pyg = [(8-l,def_shift+l-s,m,c) for l,s,m,c in zip(surface.get_losses(),surface.get_shifts(),surface.get_masks(),'RGBA')]
    pyg.sort(key=lambda x: x[1])
    offset = 0
    for size,o,m,c in pyg:
        assert o >= offset
        
        if not size: continue
        
        if o > offset:
            channels.append(ntracer.render.Channel(o - offset,0,0,0))
        channels.append(ntracer.render.Channel(
            size,
            c == 'R',
            c == 'G',
            c == 'B',
            c == 'A'))
        offset = o + size
    
    assert offset <= bs*8
        
    return channels


# "Bases" is added manually instead of using :show-inheritance: because the it
# doesn't link the base to its documentation for some reason
class PygameRenderer(ntracer.render.CallbackRenderer):
    """Bases: :py:class:`.render.CallbackRenderer`
    
    A renderer for Pygame surfaces.

    The scene is drawn onto a `pygame.Surface
    <http://www.pygame.org/docs/ref/surface.html#pygame.Surface>`_ object and
    upon completion, sends a Pygame event of type :py:attr:`ON_COMPLETE` with a
    ``source`` attribute set to the instance of the renderer, and ``surface``
    and ``scene`` attributes with the surface and scene that were passed to
    :py:meth:`begin_render`.

    Note: the renderer does not honor clipping areas or subsurface boundaries
    (it will always draw onto the entire surface) and indexed color modes (all
    8-bit color modes are indexed) are not supported.
    
    :param integer threads: The number of threads to use. If zero, the number
        of threads will equal the number of processing cores of the machine.
    
    """
    
    #: The Pygame event type that will be sent when rendering is finished.
    #:
    #: This is set to ``pygame.USEREVENT`` initially, but can be set to any
    #: value between ``pygame.USEREVENT`` and ``pygame.NUMEVENTS`` to avoid
    #: conflicts with other producers of events.
    ON_COMPLETE = pygame.USEREVENT
    
    instances = weakref.WeakSet()
    
    def __init__(self,threads=0):
        super(PygameRenderer,self).__init__(threads)
        PygameRenderer.instances.add(self)
        self.last_channels = (None,None)
    
    def begin_render(self,surface,scene):
        """Begin rendering ``scene`` onto ``dest``.

        If the renderer is already running, an exception is thrown instead.
        Upon starting, the scene will be locked for writing.

        :param pygame.Surface dest: A surface to draw onto.
        :param scene: The scene to draw.
        :type scene: :py:class:`.render.Scene`
        
        """
        def on_complete(x):
            pygame.event.post(pygame.event.Event(
                self.ON_COMPLETE,
                source=self,
                scene=scene,
                surface=surface))
        
        py_format = (surface.get_bitsize(),surface.get_masks())
        if py_format != self.last_channels[0]:
            self.last_channels = (py_format,channels_from_surface(surface))

        super(PygameRenderer,self).begin_render(
            surface.get_view() if hasattr(surface,'get_view') else surface.get_buffer(),
            ntracer.render.ImageFormat(
                surface.get_width(),
                surface.get_height(),
                self.last_channels[1],
                surface.get_pitch(),
                pygame.get_sdl_byteorder() == pygame.LIL_ENDIAN),
            scene,
            on_complete)


# When PyGame shuts down, it destroys all surface objects regardless of
# reference counts, so any threads working with surfaces need to be stopped
# first.
def _quit_func():
    for inst in PygameRenderer.instances:
        inst.abort_render()

pygame.register_quit(_quit_func)



# must work in CPython 2.7 and 3.0+

import os.path as path
import subprocess
import warnings

def get_version(base_dir):
    version = None
    try:
        # if this is a source distribution
        with open(path.join(base_dir,'PKG-INFO')) as pinfo:
            for line in pinfo:
                property,_,value = line.partition(':')
                if value and property == 'Version':
                    version = value.strip()
                    break
            else:
                assert False,"no version in PKG-INFO for some reason"
    except IOError:
        # if this is a git working tree
        try:
            version = subprocess.check_output(['git','describe','--long','--dirty'])
        except Exception:
            warnings.warn('cannot determine package version')
        else:
            # needed for Python 3
            if not isinstance(version,str):
                version = str(version,'utf-8')
            
            version = version.strip().split('-')
            del version[2] # get rid of the revision hash
            assert version[0][0] == 'v'
            version[0] = version[0][1:] # get rid of the leading "v"
            version[0] = '.'.join(version[0:2])
            del version[1]
            version = '-'.join(version)
            
    return version



"""The names :py:class:`.render.Color`, :py:class:`.render.Material`,
:py:class:`.render.Channel`, :py:class:`.render.ImageFormat`,
:py:class:`.render.CallbackRenderer`, :py:class:`.render.BlockingRenderer`,
:py:class:`.wrapper.NTracer`, :py:data:`.wrapper.CUBE` and
:py:data:`.wrapper.SPHERE` are imported here for convenience."""


from ntracer.render import Color,Material,Channel,ImageFormat,CallbackRenderer,BlockingRenderer
from ntracer.wrapper import NTracer,CUBE,SPHERE




from . import render
from . import wrapper

class FileFormatError(Exception):
    def __init__(self):
        super(FileFormatError,self).__init__("not a valid wavefront file")


def index1(x):
    return x-1 if x >= 0 else x

def load_obj(file,nt=None):
    if nt is None:
        nt = wrapper.NTracer(3)
    elif nt.dimension != 3:
        raise ValueError('Wavefront .obj files only support 3-dimensional geometry')
    
    m = render.Material((1,1,1))
        
    vertices = []
    triangles = []
    
    with open(file,'r') as input:
        for line in input:
            parts = line.split()
            if parts[0] == 'v':
                try:
                    coords = [float(p) for p in parts[1:4]]
                except ValueError:
                    raise FileFormatError()
                vertices.append(nt.Vector(coords))
            elif parts[0] == 'f':
                try:
                    # the neat thing here is we don't have to do anything to
                    # support OBJ's relative indices, since they are negative
                    # indices that work the same as they do for Python lists
                    coords = [vertices[index1(int(i.partition('/')[0],10))] for i in parts[1:]]
                except (ValueError,IndexError):
                    raise FileFormatError()
            
                for i in range(1,len(coords)-1):
                    triangles.append(nt.TrianglePrototype([coords[0],coords[i],coords[i+1]],m))
                
    return triangles



#!/usr/bin/python

from __future__ import unicode_literals,print_function

import math
import fractions
import pygame
import argparse
import os.path
import sys
import subprocess
import time
from itertools import combinations,islice
from ntracer import NTracer,Material,ImageFormat,Channel,BlockingRenderer,CUBE
from ntracer.pygame_render import PygameRenderer


ROT_SENSITIVITY = 0.005
WHEEL_INCREMENT = 8


def excepthook(type,value,traceback):
    if isinstance(value,Exception):
        print('error: '+str(value),file=sys.stderr)
    else:
        sys.__excepthook__(type,value,traceback)
    
sys.excepthook = excepthook


def schlafli_component(x):
    x = x.partition('/')
    p = int(x[0],10)
    if p < 3: raise argparse.ArgumentTypeError('a component cannot be less than 3')
    
    if not x[2]: return fractions.Fraction(p)
    
    s = int(x[2],10)
    if s < 1: raise argparse.ArgumentTypeError('for component p/q: q cannot be less than 1')
    if s >= p: raise argparse.ArgumentTypeError('for component p/q: q must be less than p')
    if fractions.gcd(s,p) != 1: raise argparse.ArgumentTypeError('for component p/q: p and q must be co-prime')
    return fractions.Fraction(p,s)

def positive_int(x):
    x = int(x,10)
    if x < 1: raise argparse.ArgumentTypeError('a positive number is required')
    return x

def screen_size(x):
    w,_,h = x.partition('x')
    w = int(w,10)
    h = int(h,10)
    if w < 1 or h < 1: raise argparse.ArgumentTypeError('invalid screen size')
    return w,h

def fov_type(x):
    x = float(x)
    if x <= 0 or x >= 180: raise argparse.ArgumentTypeError('fov must be between 0 and 180 degrees')
    return x/180*math.pi

parser = argparse.ArgumentParser(
    description='Display a regular polytope given its Schl\u00e4fli symbol.')
parser.add_argument('schlafli',metavar='N',type=schlafli_component,nargs='+',help='the Schl\u00e4fli symbol components')
parser.add_argument('-o','--output',metavar='PATH',help='save an animation to PATH instead of displaying the polytope')
parser.add_argument('-t','--type',metavar='TYPE',default='h264',
    help='Specifies output type when --output is used. If TYPE is "png", the '+
        'output is a series of PNG images. For any other value, it is used '+
        'as the video codec for ffmpeg.')
parser.add_argument('-f','--frames',metavar='F',type=positive_int,default=160,help='when creating an animation or benchmarking, the number of frames to render')
parser.add_argument('-s','--screen',metavar='WIDTHxHEIGHT',type=screen_size,default=(800,600),help='screen size')
parser.add_argument('-a','--fov',metavar='FOV',type=fov_type,default=0.8,help='field of vision in degrees')
parser.add_argument('-d','--cam-dist',metavar='DIST',type=float,default=4,
    help='How far the view-port is from the center of the polytope. The '+
        'value is a multiple of the outer raidius of the polytope.')
parser.add_argument('--benchmark',action='store_true',help='measure the speed of rendering the scene')
args = parser.parse_args()


material = Material((1,0.5,0.5))
nt = NTracer(max(len(args.schlafli)+1,3))


def higher_dihedral_supplement(schlafli,ds):
    a = math.pi*schlafli.denominator/schlafli.numerator
    return 2*math.asin(math.sin(math.acos(1/(math.tan(ds/2)*math.tan(a))))*math.sin(a))

def almost_equal(a,b,threshold=0.1):
    return (a-b).absolute() < threshold

def radial_vector(angle):
    return nt.Vector.axis(0,math.sin(angle)) + nt.Vector.axis(1,math.cos(angle))


class Instance:
    def __init__(self,shape,position,orientation=nt.Matrix.identity()):
        self.shape = shape
        self.position = position
        self.orientation = orientation
        self.inv_orientation = orientation.inverse()
    
    def translated(self,position=nt.Vector(),orientation=nt.Matrix.identity()):
        return (
            position + (orientation * self.position),
            orientation * self.orientation)
    
    def tesselate(self,*args):
        return self.shape.tesselate(*self.translated(*args))
    
    def tesselate_inner(self,*args):
        return self.shape.tesselate_inner(*self.translated(*args))
    
    def any_point(self,*args):
        return self.shape.any_point(*self.translated(*args))
    
    def contains(self,p):
        return self.shape.contains(self.inv_orientation * (p - self.position))


def star_component(x):
    return (x.numerator - 1) > x.denominator > 1

class LineSegment:
    star = False
    
    def __init__(self,index,convex_ds,polygon):
        self.index = index
        self.p = polygon
        self.position = radial_vector(index*convex_ds)
    
    def tesselate(self,position,orientation):
        return [
            orientation*self.p.base_points[self.index-1]+position,
            orientation*self.p.base_points[self.index]+position]


class Polygon:
    apothem = 1
    
    def __init__(self,schlafli):
        self.star = star_component(schlafli)
        convex_ds = 2 * math.pi / schlafli.numerator
        self.dihedral_s = convex_ds * schlafli.denominator
        self.parts = [LineSegment(i,convex_ds,self) for i in range(schlafli.numerator)]
        
        self._circumradius = 1/math.cos(convex_ds/2)
        self.base_points = [self._circumradius * radial_vector((i+0.5) * convex_ds) for i in range(schlafli.numerator)]
        
        if self.star:
            self._circumradius = math.tan(convex_ds)*math.tan(convex_ds/2) + 1
            self.outer_points = [self._circumradius * radial_vector(i * convex_ds) for i in range(schlafli.numerator)]

    def points(self,position,orientation,pset=None):
        if pset is None: pset = self.base_points
        return (orientation * bp + position for bp in pset)
    
    def tesselate_inner(self,position,orientation):
        points = list(self.points(position,orientation))
        r = [points[0:3]]
        for i in range(len(points)-3):
            r.append([points[0],points[i+2],points[i+3]])
        return r
    
    def tesselate(self,position,orientation):
        if not self.star:
            return self.tesselate_inner(position,orientation)
        
        points = list(self.points(position,orientation))
        opoints = list(self.points(position,orientation,self.outer_points))
        return [[opoints[i],points[i-1],points[i]] for i in range(len(points))]
    
    def any_point(self,position,orientation):
        return next(self.points(position,orientation))
    
    def contains(self,p):
        return any(almost_equal(p,test_p) for test_p in self.base_points)
    
    def hull(self,position=nt.Vector(),orientation=nt.Matrix.identity()):
        tris = [nt.TrianglePrototype(tri,material) for tri in self.tesselate_inner(position,orientation)]
        if self.star: tris.extend(nt.TrianglePrototype(tri,material) for tri in
            self.tesselate(position,orientation))
        return tris
    
    def circumradius(self):
        return self._circumradius
    
    def circumradius_square(self):
        return self._circumradius*self._circumradius
    
    def line_apothem_square(self):
        return 1


class Plane:
    def __init__(self,nt,position):
        self.normal = position.unit()
        self.d = -position.absolute()
        self._dot = nt.dot
    
    def distance(self,point):
        return self._dot(point,self.normal) + self.d

class Line:
    def __init__(self,nt,p0,v,planes,outer=False):
        self.p0 = p0
        self.v = v
        self.planes = set(planes)
        self.outer = outer
        self._dot = nt.dot
    
    def point_at(self,t):
        return self.p0 + self.v*t
    
    def dist_square(self,point):
        a = point - self.p0
        b = self._dot(a,self.v)
        return a.square() - b*b/self.v.square()
    
    def __repr__(self):
        return 'Line({0!r},{1!r})'.format(self.p0,self.v)


def plane_point_intersection(nt,planes):
    assert nt.dimension == len(planes)
    try:
        return nt.Matrix(p.normal for p in planes).inverse()*nt.Vector(-p.d for p in planes)
    except ValueError:
        return None

def plane_line_intersection(nt,planes):
    assert nt.dimension - 1 == len(planes)
    v = nt.cross(p.normal for p in planes).unit()
    return Line(
        nt,
        nt.Matrix([p.normal for p in planes] + [v]).inverse() * nt.Vector([-p.d for p in planes] + [0]),
        v,
        planes)

def line_intersection(nt,l1,l2):
    d = nt.dot(l1.v,l2.v)
    denom = 1 - d*d
    if not denom: return None
    id = 1/denom
    a = nt.dot(l2.p0 - l1.p0,l1.v)
    b = nt.dot(l1.p0 - l2.p0,l2.v)
    t1 = id*(a + d*b)
    t2 = id*(d*a + b)
    p1 = l1.point_at(t1)
    p2 = l2.point_at(t2)
    if abs(p1-p2) > 0.01: return None
    return (p1 + p2) * 0.5, t1, t2


class Node:
    def __init__(self,pos,planes,outer,alive=True):
        self.pos = pos
        self.planes = planes
        self.outer = outer
        self.neighbors = set() if alive else None
    
    def detach(self):
        for n in self.neighbors:
            n.neighbors.remove(self)
        self.neighbors = None
    
    @property
    def dead(self):
        return self.neighbors is None
    
    def find_cycles(self,length,sequence=None,exclude=None):
        if sequence is None: sequence = [self]
        if len(sequence) < length:
            exclude = exclude.copy() if exclude is not None else set([self])
            for n in self.neighbors:
                if n not in exclude:
                    exclude.add(n)
                    for r in n.find_cycles(length,sequence + [n],exclude):
                        yield r
        else:
            for n in self.neighbors:
                if n is sequence[0] and n.planes.intersection(*(sequence[i].planes for i in range(1,len(sequence)))):
                    yield sequence

def join(a,b):
    if not (a.dead or b.dead):
        a.neighbors.add(b)
        b.neighbors.add(a)

class FuzzyGraph:
    def __init__(self):
        self.nodes = []
    
    def add(self,pos,planes,outer):
        for n in self.nodes:
            if almost_equal(n.pos,pos,0.01):
                n.planes |= planes
                return n
        n = Node(pos,planes,outer)
        self.nodes.append(n)
        return n
    
    def remove_at(self,i):
        self.nodes[i].detach()
        if i+1 != len(self.nodes):
            self.nodes[i] = self.nodes[-1]
        del self.nodes[-1]
    
    def remove(self,pos):
        if isinstance(pos,Node):
            if not pos.dead:
                self.remove_at(self.nodes.index(pos))
        else:
            for i,n in enumerate(self.nodes):
                if almost_equal(n.pos,pos,0.01):
                    self.remove_at(i)
                    break


# Cells are enlarged ever so slightly to prevent the view frustum from being
# wedged exactly between two adjacent primitives, which, do to limited
# precision, can cause that volume to appear to vanish.
fuzz_scale = nt.Matrix.scale(1.00001)
class PolyTope:
    def __init__(self,dimension,schlafli,dihedral_s,face_apothem):
        self.dimension = dimension
        self.schlafli = schlafli
        self.dihedral_s = dihedral_s
        self.apothem = math.tan((math.pi - dihedral_s)/2) * face_apothem
        self.star = star_component(schlafli)
        self.parts = []
    
    @property
    def facet(self):
        return self.parts[0].shape
        
    def propogate_faces(self,potentials):
        new_p = []
        
        for instance,p in potentials:
            dir = (instance.orientation * p.position).unit()
            
            reflect = nt.Matrix.reflection(dir)

            turn = nt.Matrix.rotation(
                instance.position.unit(),
                dir,
                self.dihedral_s)
            
            new_p += self.add_face(Instance(
                instance.shape,
                turn * instance.position,
                fuzz_scale * turn * reflect * instance.orientation))
        return new_p
    
    def add_face(self,instance):
        for p in self.parts:
            if almost_equal(instance.position,p.position): return []

        self.parts.append(instance)
        
        return [(instance,p) for p in instance.shape.parts]
    
    def star_tesselation(self):
        t = getattr(self,'_star_tesselation',None)
        if t is None:
            co_nt = NTracer(self.dimension)
            lines = []
            planes = [Plane(co_nt,co_nt.Vector(islice(part.position,co_nt.dimension))) for part in self.parts]
            las = self.line_apothem_square()
            for pgroup in combinations(planes,co_nt.dimension-1):
                try:
                    line = plane_line_intersection(co_nt,pgroup)
                except ValueError:
                    pass
                else:
                    if line:
                        for lineb in lines:
                            if almost_equal(line.p0,lineb.p0) and almost_equal(line.v,lineb.v):
                                lineb.planes |= line.planes
                                break
                        else:
                            outer_dist = line.dist_square(co_nt.Vector()) - las
                            if outer_dist < 0.1:
                                line.outer = outer_dist > -0.1
                                lines.append(line)

            pmap = {}
            for line in lines:
                pmap[line] = {}

            graph = FuzzyGraph()
            maxr = self.circumradius_square() + 0.1
            for l1,l2 in combinations(lines,2):
                inter = line_intersection(co_nt,l1,l2)
                if inter and inter[0].square() < maxr:
                    n = graph.add(inter[0],l1.planes | l2.planes,l1.outer or l2.outer)
                    pmap[l1][n] = inter[1]
                    pmap[l2][n] = inter[2]

            for line,poss in pmap.items():
                if len(poss) == 0: continue
                if len(poss) == 1:
                    graph.remove(poss[0])
                    continue
                
                poss = sorted(poss.items(),key=(lambda x: x[1]))
                
                if line.outer:
                    for i in range(len(poss)-1):
                        join(poss[i][0],poss[i+1][0])
                elif len(poss) == 2:
                    join(poss[0][0],poss[1][0])
                elif len(poss) > 3:
                    for i in range(2,len(poss)-2):
                        graph.remove(poss[i][0])

                    join(poss[0][0],poss[1][0])
                    join(poss[-1][0],poss[-2][0])
            
            t = []
            self._star_tesselation = t
            for n in islice(graph.nodes,0,len(graph.nodes)-co_nt.dimension):
                for cycle in n.find_cycles(co_nt.dimension):
                    t.append([nt.Vector(tuple(x.pos) + (0,) * (nt.dimension-co_nt.dimension)) for x in cycle] + [nt.Vector()])
                n.detach()

        return t
    
    def tesselate(self,position,orientation):
        if self.star or self.facet.star:
            return [[orientation * p + position for p in tri] for tri in self.star_tesselation()]

        return self.tesselate_inner(position,orientation)
    
    def tesselate_inner(self,position,orientation):
        tris = []
        point1 = self.parts[0].any_point(position,orientation)
            
        inv_orientation = orientation.inverse()
        for part in self.parts[1:]:
            if not part.contains(inv_orientation * (point1 - position)):
                new_t = part.tesselate(position,orientation)
                for t in new_t: t.append(point1)
                tris += new_t
            
        return tris
    
    def hull(self,position=nt.Vector(),orientation=nt.Matrix.identity()):
        tris = []
        for p in self.parts:
            tris += p.tesselate(position,orientation)
        return [nt.TrianglePrototype(tri,material) for tri in tris]
    
    def any_point(self,position,orientation):
        return self.parts[0].any_point(position,orientation)
    
    def contains(self,p):
        return any(part.contains(p) for part in self.parts)
    
    def circumradius_square(self):
        return self.apothem*self.apothem + self.facet.circumradius_square()
    
    def circumradius(self):
        return math.sqrt(self.circumradius_square())
    
    def line_apothem_square(self):
        return self.apothem*self.apothem + self.facet.line_apothem_square()


def compose(part,order,schlafli):
    if schlafli.numerator * (math.pi - part.dihedral_s) >= math.pi * 2 * schlafli.denominator:
        exit("Component #{0} ({1}) is invalid because the angles of the parts add up to 360\u00b0 or\nmore and thus can't be folded inward".format(order,schlafli))
    higher = PolyTope(
        order+1,
        schlafli,
        higher_dihedral_supplement(schlafli,part.dihedral_s),
        part.apothem)
    potentials = higher.add_face(Instance(part,nt.Vector.axis(order,higher.apothem)))
    while potentials:
        potentials = higher.propogate_faces(potentials)
    return higher


jitter = nt.Vector((0,0,0) + (0.0001,) * (nt.dimension-3))
def process_movement():
    global x_move, y_move, w_move
    
    if x_move or y_move or w_move:
        h = math.sqrt(x_move*x_move + y_move*y_move + w_move*w_move)
        a2 = camera.axes[0]*(x_move/h) + camera.axes[1]*(-y_move/h)
        if w_move: a2 += camera.axes[3] * (w_move / h)
        camera.transform(nt.Matrix.rotation(
            camera.axes[2],
            a2,
            h * ROT_SENSITIVITY))
        camera.normalize()
        camera.origin = camera.axes[2] * cam_distance + jitter
        scene.set_camera(camera)
        
        x_move = 0
        y_move = 0
        w_move = 0
        
        run()


def run():
    global running
    
    running = True
    render.begin_render(screen,scene)


try:
    timer = time.perf_counter
except AttributeError:
    timer = time.clock
    if args.benchmark and not sys.platform.startswith('win'):
        print('''warning: on multi-core systems, Python\'s high-resolution timer may combine
time spent on all cores, making the reported time spent rendering, much higher
than the actual time''',file=sys.stderr)


class RotatingCamera(object):
    incr = 2 * math.pi / args.frames
    h = 1/math.sqrt(nt.dimension-1)
    
    _timer = staticmethod(timer if args.benchmark else (lambda: 0))
    
    def __enter__(self):
        self.frame = 0
        self.total_time = 0
        return self
    
    def __exit__(self,type,value,tb):
        if type is None and self.total_time:
            print('''rendered {0} frame(s) in {1} seconds
time per frame: {2} seconds
frames per second: {3}'''.format(self.frame,self.total_time,self.total_time/self.frame,self.frame/self.total_time))
    
    def start_timer(self):
        self.t = self._timer()
    
    def end_timer(self):
        self.total_time += self._timer() - self.t
    
    def advance_camera(self):
        self.frame += 1
        if self.frame >= args.frames: return False
            
        a2 = camera.axes[0]*self.h + camera.axes[1]*self.h
        for i in range(nt.dimension-3): a2 += camera.axes[i+3]*self.h
        camera.transform(nt.Matrix.rotation(camera.axes[2],a2,self.incr))
        camera.normalize()
        camera.origin = camera.axes[2] * cam_distance
        scene.set_camera(camera)
        
        return True


if nt.dimension >= 3 and args.schlafli[0] == 4 and all(c == 3 for c in args.schlafli[1:]):
    cam_distance = -math.sqrt(nt.dimension) * args.cam_dist
    scene = nt.BoxScene()
else:
    print('building geometry...')
    timing = timer()
    
    p = Polygon(args.schlafli[0])
    for i,s in enumerate(args.schlafli[1:]):
        p = compose(p,i+2,s)

    hull = p.hull()
    timing = timer() - timing
    print('done in {0} seconds'.format(timing))

    cam_distance = -math.sqrt(p.circumradius_square()) * args.cam_dist

    print('partitioning scene...')
    timing = timer()
    scene = nt.build_composite_scene(hull)
    timing = timer() - timing
    print('done in {0} seconds'.format(timing))
    
    del p
    del hull

camera = nt.Camera()
camera.translate(nt.Vector.axis(2,cam_distance) + jitter)
scene.set_camera(camera)
scene.set_fov(args.fov)


if args.output is not None:
    if args.type != 'png':
        render = BlockingRenderer()
        format = ImageFormat(
            args.screen[0],
            args.screen[1],
            [Channel(16,1,0,0),
             Channel(16,0,1,0),
             Channel(16,0,0,1)])
        
        surf = bytearray(args.screen[0]*args.screen[1]*format.bytes_per_pixel)
        
        pipe = subprocess.Popen(['ffmpeg',
                '-y',
                '-f','rawvideo',
                '-vcodec','rawvideo',
                '-s','{0}x{1}'.format(*args.screen),
                '-pix_fmt','rgb48be',
                '-r','60',
                '-i','-',
                '-an',
                '-vcodec',args.type,
                '-crf','10',
                args.output],
            stdin=subprocess.PIPE)
        
        try:
            with RotatingCamera() as rc:
                while True:
                    rc.start_timer()
                    render.render(surf,format,scene)
                    rc.end_timer()

                    print(surf,file=pipe.stdin,sep='',end='')
                    
                    if not rc.advance_camera(): break

        finally:
            pipe.stdin.close()
            r = pipe.wait()
        sys.exit(r)


    pygame.display.init()
    render = PygameRenderer()
    surf = pygame.Surface(args.screen,depth=24)

    def announce_frame(frame):
        print('drawing frame {0}/{1}'.format(frame+1,args.frames))
    
    with RotatingCamera() as rc:
        announce_frame(0)
        rc.start_timer()
        render.begin_render(surf,scene)
        while True:
            e = pygame.event.wait()
            if e.type == pygame.USEREVENT:
                rc.end_timer()
                
                pygame.image.save(
                    surf,
                    os.path.join(args.output,'frame{0:04}.png'.format(rc.frame)))

                if not rc.advance_camera(): break
                
                announce_frame(rc.frame)
                rc.start_timer()
                render.begin_render(surf,scene)
            elif e.type == pygame.QUIT:
                render.abort_render()
                break

else:
    pygame.display.init()
    render = PygameRenderer()
    screen = pygame.display.set_mode(args.screen)
    
    if args.benchmark:
        with RotatingCamera() as rc:
            rc.start_timer()
            render.begin_render(screen,scene)
            while True:
                e = pygame.event.wait()
                if e.type == pygame.USEREVENT:
                    rc.end_timer()
                    
                    pygame.display.flip()

                    if not rc.advance_camera(): break
                    
                    rc.start_timer()
                    render.begin_render(screen,scene)
                elif e.type == pygame.QUIT:
                    render.abort_render()
                    break
    else:
        running = False
        run()

        x_move = 0
        y_move = 0
        w_move = 0

        while True:
            e = pygame.event.wait()
            if e.type == pygame.MOUSEMOTION:
                if e.buttons[0]:
                    x_move += e.rel[0]
                    y_move += e.rel[1]
                    if not running:
                        process_movement()
            elif e.type == pygame.MOUSEBUTTONDOWN:
                if nt.dimension > 3:
                    if e.button == 4 or e.button == 5:
                        if e.button == 4:
                            w_move += WHEEL_INCREMENT
                        else:
                            w_move -= WHEEL_INCREMENT
                        if not running:
                            process_movement()
            elif e.type == pygame.USEREVENT:
                running = False
                pygame.display.flip()
                process_movement()
            elif e.type == pygame.KEYDOWN:
                if e.key == pygame.K_c:
                    x,y = pygame.mouse.get_pos()
                    fovI = (2 * math.tan(scene.fov/2)) / screen.get_width()

                    print(camera.origin)
                    print((camera.axes[2] + camera.axes[0] * (fovI * (x - screen.get_width()/2)) - camera.axes[1] * (fovI * (y - screen.get_height()/2))).unit())
            elif e.type == pygame.QUIT:
                render.abort_render()
                break




import importlib
import weakref

import ntracer.render

CUBE = 1
SPHERE = 2


def vector_wrapper(mod,dim):
    base = mod.Vector
    class Vector(base):
        __slots__ = ()
        
        def __new__(cls,*values):
            if len(values) > 1: return base.__new__(cls,dim,values)
            return base.__new__(cls,dim,*values)
        
        @staticmethod
        def axis(axis,length=1):
            return base.axis(dim,axis,length)
        
    return Vector

def matrix_wrapper(mod,dim):
    base = mod.Matrix
    class Matrix(base):
        __slots__ = ()
        
        def __new__(cls,*values):
            if len(values) > 1: return base.__new__(cls,dim,values)
            return base.__new__(cls,dim,*values)
        
        @staticmethod
        def scale(factor):
            if isinstance(factor,mod.Vector):
                return base.scale(factor)
            return base.scale(dim,factor)
        
        @staticmethod
        def identity():
            return base.identity(dim)
        
    return Matrix

def camera_wrapper(mod,dim):
    base = mod.Camera
    class Camera(base):
        def __new__(cls):
            return base.__new__(cls,dim)

        def __init__(self):
            base.__init__(self,dim)
            
    return Camera

def boxscene_wrapper(mod,dim):
    base = mod.BoxScene
    class BoxScene(base):
        def __new__(cls):
            return base.__new__(cls,dim)
        
    return BoxScene

def aabb_wrapper(mod,dim):
    base = mod.AABB
    class AABB(base):
        def __new__(cls,*args,**kwds):
            return base.__new__(cls,dim,*args,**kwds)

    return AABB


class NTracer(object):
    """A helper class that simplifies the creation of multiple objects with
    the same dimension.
    
    An instance contains all the same values that :py:mod:`.tracern` has,
    except every class that has a method with a ``dimension`` parameter is
    subclassed so that parameter is filled in automatically. Additionally, the
    constructors for :py:class:`.tracern.Vector` and
    :py:class:`.tracern.Matrix` are modified to accept multiple parameters as
    an alternative to a sequence of values. e.g.:
    :code:`NTracer(3).Vector([1,2,3])` can also be written as
    :code:`NTracer(3).Vector(1,2,3)`.
    
    An instance of ``NTracer`` imports from a specialized (read: faster)
    tracer\ *{dimension}* version if it exists, and falls back to
    :py:mod:`.tracern` otherwise.
    
    Instances of ``NTracer`` are cached, so invoking ``NTracer`` while another
    instance with the same dimension already exists, will simply give you the
    first instance (unless ``force_generic`` is given a value of ``True``).
    
    :param integer dimension: The value that will be automatically given to any
        function/method/constructor that requires a ``dimension`` parameter.
    :param boolean force_generic: If ``True``, :py:mod:`.tracern` will be used
        even if a specialized version exists. This is mainly for testing.
    """
    
    _cache = weakref.WeakValueDictionary()
    
    def __new__(cls,dimension,force_generic=False):
        if not force_generic:
            obj = NTracer._cache.get(dimension)
            if obj is not None: return obj
        
        obj = object.__new__(cls)
        
        if force_generic:
            mod = importlib.import_module('ntracer.tracern')
        else:
            mod = ntracer.render.get_optimized_tracern(dimension)
        
        obj.dimension = dimension
        obj.Vector = vector_wrapper(mod,dimension)
        obj.Matrix = matrix_wrapper(mod,dimension)
        obj.Camera = camera_wrapper(mod,dimension)
        obj.BoxScene = boxscene_wrapper(mod,dimension)
        obj.AABB = aabb_wrapper(mod,dimension)
        
        for n in [
            'CompositeScene',
            'KDNode',
            'KDLeaf',
            'KDBranch',
            'Primitive',
            'PrimitivePrototype',
            'Solid',
            'SolidPrototype',
            'Triangle',
            'TrianglePrototype',
            'PointLight',
            'GlobalLight',
            'dot',
            'cross',
            'build_kdtree',
            'build_composite_scene']:
            setattr(obj,n,getattr(mod,n))
        
        if not force_generic:
            NTracer._cache[dimension] = obj
            
        return obj

